

#include "trithorpe.h"

long tab::countCtors;
long tab::countDtors;
int verbose;
Sayer say;

tri vec::const_zero;


void SymToAsciiAdd( tri& v, const tri& x ) {
    MUST( x.isSym() );
    word w= x.peek();
    u1 a= w>>24;
    u1 b= w>>16;
    u1 c= w>> 8;
    u1 d= w    ;
    v << '`';
    if (a!='@') v<<(char)a;
    if (b!= 0 ) v<<(char)b;
    if (c!= 0 ) v<<(char)c;
    if (d!= 0 ) v<<(char)d;
    v << '\'';
}
void IntToAsciiAdd( tri& v, int x ) {
    if (x<0) { v.aAdd( '-' ); x= -x; }
    if (x<10) { v.aAdd( '0'+x ); return; }
    IntToAsciiAdd ( v, x/10 );
    v.aAdd( '0'+(x%10) );
}

static void tabelize(tri& t) {
	if ( ! t.isTab() ) {
	    tri tmp= t;
	    t.setNewTab();
	    if (tmp != NIL) t << tmp;  // NIL is exception; behaves like empty vec
	}
}

static void flatten(tri& z, tri a, int depth) {

	if (depth<1) {
	    z.aAdd( '#' );

	} else if ( a.isTab() ) {

	    int n= a.aSize();
	    if ( n>0 && IsPrintableAsciiObj(a) ) {
		//?--IntToAsciiAdd( z, a.asInt() );
		for (int i=0; i<n; i++) {
		    z.aAdd( a.aGet(i) );
		}
	    } else {
		z.aAdd( (int)'{' );
		for (int i=0; i<n; i++) {
		    flatten(z, a.aGet(i), depth-1);
		}
		z.aAdd( (int)'}' );
	    }
	} else if ( a.isInt() ) {
	    z.aAdd( a );
	} else if ( a.isSym() ) {
	    SymToAsciiAdd( z, a );
	} else {
	    abort();
	}
}
tri& tri::operator << (const tri& a) {
    tabelize( *this );
    // Special case at outermost level for INT
    if ( a.isInt() ) 
	IntToAsciiAdd( *this, a.asInt() );
    else
	flatten( *this, a, 4 );
    return *this;
}

tri& tri::operator << (word a) { return *this << (int)a; }
tri& tri::operator << (int a) {
	tabelize(*this);
	IntToAsciiAdd( *this, a );
	return *this;
}
tri& tri::operator << (unsigned char a) {
	tabelize(*this);
	this->aAdd( (int) a);
	return *this;
}
tri& tri::operator << (char a) {
	tabelize(*this);
	this->aAdd( (int) a);
	return *this;
}
tri& tri::operator << (const char* a) {
	tabelize(*this);
	for (int i=0; a[i]; i++) 
	    this->aAdd( (int) a[i] );
	return *this;
}
tri& tri::operator << (const tab& a) {
	tabelize(*this);
	return *this << "(?)";
}
tri& tri::operator << (const vec& a) {
	tabelize(*this);
	return *this << "(?)";
}
tri& tri::operator << (const TriTerp& a) {
	tabelize(*this);
	return *this << "(?)";
}
tri& tri::operator << (class Expr* a) {
	tabelize(*this);
	return *this << "(?)";
}


tri::tri(const char* tcodeScript)
{
	guts= NIL_GUTS; // temporary guts.

	setTCodeArrayFromString(tcodeScript, NULL, true);
}

tri tri::NewCodVec(const char* tcodeScript)
{
	tri x;
	x.setTCodeArrayFromString(tcodeScript, NULL, false);
	return x;
}

tri tri::NewDeepCodVec(const char* tcodeScript)
{
	tri x;
	x.setTCodeArrayFromString(tcodeScript, NULL, true);
	return x;
}

tri & tri::setSym(char a, char b, char c, char d)
{
    unlink();
    word x = (u1(a) << 24) | (u1(b) << 16) | (u1(c) << 8) | u1(d);

    EQ(SYM, x & HI);

    guts = x;
    return *this;
}

tri & tri::setSym(word x)
{
    unlink();

    EQ(SYM, x & HI);

    guts = (x & LO | SYM & HI);
    return *this;
}

#ifdef DEBUG
void tri::to( FILE* f ) const
{
	char buf[99999];
	toString(buf);
	fprintf( f, "%s", buf );
}

void
tri::toString( char buf[] ) const
{
    switch (isWhat()) {
    case INT:	sprintf( buf, "%d", guts ); break;
    case SYM:	{
	char* s= buf;
	u1 x= guts>>24;
	if (x!='@') *s++= x;
	x= guts>>16;
	if (x) *s++= x;
	x= guts>>8;
	if (x) *s++= x;
	x= guts;
	if (x) *s++= x;
	*s = 0;
    }
    break;
    case TAB:	{
	tab* p= asTab();
	if (p->meta && p->meta->tag != NIL && p->meta->tag.isSym() ) {
		char buf2[99999];
		p->meta->tag.toString(buf2);
		sprintf( buf, "<%s>", buf2 );
	} else {
		char* cp= StrDupIfPrintableAsciiObj( *this );
		if (cp) {
		    sprintf( buf, "<%d:%d:0x%x:\"%s\">", p->arr.size(), p->dict.size(), (int)p, cp );
		    delete [] cp;
		} else {
		    sprintf( buf, "<%d:%d:0x%x>", p->arr.size(), p->dict.size(), (int)p );
		}
	}
    }
    break;
    default:	sprintf( buf, "?%x?", guts ); break;
    }
}
#endif


tab::tab(tri * setMe)
    : refs(0)
#ifdef MAGIC
    , magic(MAGIC)
#endif
    , meta(0)
{
    ++countCtors;
    setMe->setTab(this);
}

tab::~tab()
{
    checkMagic();
    MUST( refs==0 );
    ++countDtors;
    if (meta && meta->finalizer) meta->finalizer(this);
}

void vec::do_grow(int goal)
{
    if (cap == 0) {
	cap = goal < 8 ? 8 : goal;
	d = new tri[cap];
    } else {
	int n = cap * 2;
	while (n < goal) {
	    n = n * 2;
	}
	tri *z = new tri[n];	// make the new
	for (int i = 0; i < cap; i++) {
	    z[i] = d[i];
	    d[i].setInt(0);
	}
	cap = n;
	delete[]d;		// trash the old
	d = z;
    }
}

vec::~vec()
{
    if (d) {
	delete [] d;
	d = 0;
    }
    cap = 0;
    siz = 0;
}

int StrLen(const char* s) {
    for (int i=0;  ; i++) if (s[i]=='\0') return i;
}
int StrSum(const char* s) {
    int z= 0;
    for (int i=0;  ; i++) if (s[i]=='\0') return z; else z += (unsigned char)s[i];
}
char* StrDup (const char* s) {
    int n= StrLen(s);
    char* cp= new char [n+1];
    for (int i=0; i<=n; i++) cp[i]= s[i];
    return cp;
}

int StrCmp(const char* a, const char* b) {
    for (int i=0;  ; i++) {
	if ( a[i]=='\0' && b[i]=='\0' ) return 0;
	if ( a[i] < b[i] ) return -1;
	if ( a[i] > b[i] ) return  1;
    }
}

#if 0
InternStr::InternStr(tri _obj, const char* _str, InternStr* _next)
  : obj( _obj )
  , str( _str )
  , next( _next )
{}

InternStr* InternTable::intern(const char* s) {
	int h= StrSum(s);		// hash
	int b= h & INTERN_MASK;		// bucket
	InternStr* p= bucket[b];	// chain
	while (p) {
		if (StrEq( s, p->str)) return p;
		p= p->next;
	}

	// not found -- make one.

	p= new InternStr( Str(s), StrDup(s), bucket[b] );
	bucket[b]= p;
	return p;
}
#endif
	
tri Str(const char* s) {  // construct new String tri
	tri z;
	z.setNewTab();
	while (*s) {
		z.aAdd( tri( (int)*s++ ) );
	}
	return z;
}
bool IsPrintableAsciiObj(tri x) {
    if ( ! x.isTab() ) return 0;
    int n= x.aSize();
    for (int i=0; i<n ; i++) {
	tri e= x.aGet(i);
        if ( ! e.isInt() ) return 0;
	int c= e.asInt();
        if ( c < 32 || c > 126 ) return 0;
    }
    return 1;
}
char* StrDupIfPrintableAsciiObj(tri x) {
    if (IsPrintableAsciiObj(x)) {
	int n= x.aSize();
	char* z= new char [ n+1 ];
	for (int i=0; i<n ; i++) {
	    z[i] = x.aGet(i).asInt();
	}
	z[n]= '\0';
	return z;
    } else {
	return NULL;
    }
}

#ifdef MAGIC
void tab::checkMagic() const
{
    //fprintf(stderr, " {check 0x%x} ", (int)this );
    if ( magic != MAGIC ) {
	say << "BAD TAB MAGIC AT " << (int)this;
	Bad();
	abort();
    }
    if ( arr.size() & 0xFFF00000 ) {
	say << "BAD arr size AT " << (int)this;
	Bad();
	abort();
    }
    if ( dict.size() & 0xFFF00000 ) {
	say << "BAD dict size AT " << (int)this;
	Bad();
	abort();
    }
}
#endif

void CheckDeep(tri x, int depth) {
    if (depth<1) return;

    if (x.isTab()) {

        int n=x.aSize();
        for (int i=0; i<n; i++) {
            tri z= x.aGet(i);
            if (z.isTab()) {
                z.asTab()->checkMagic();
                CheckDeep( z, depth-1 );
            }
        }

    }
}

void deleteSimpleMeta(tab* p) {
    if (p && p->meta) delete p->meta;
}

tri& tri::setTag(tri tag) {
    tab* p= asTab();
    EQ( 0, (int) p->meta );
    p->meta = new Meta();
    p->meta->tag= tag;
    p->meta->finalizer= &deleteSimpleMeta;
    return *this;
}

tri tri::recall(const tri & index) const
{
    if (isTab()) {
	return asTab()->recall(index);
    } else {
	CANT("tri::recall on not a Tab");
	return tri();
    }
}
tri tab::recall(const tri & tindex) const
{
    return recall(tindex.peek());
}

tri tab::recall(word windex) const 
{
    switch (tri::isWhat(windex)) {
    case tri::INT:{
	    int ix = tri::asInt(windex);
	    //vec& hack= * (vec*) &arr; // nonconst hack
	    return (ix < arr.size())? arr[ix] : tri();
	}
    case tri::SYM:{
	    int sz = dict.size();
	    for (int i = 0; i < sz; i += 2) {
		if (windex == dict[i].peek()) {
		    return dict[i + 1];
		}
	    }
	    return tri();
	}
    }
    CANT("tab::recall() index nor INT nor SYM");
    return tri();
}



void tab::store(const tri & tindex, const tri & value)
{
    store(tindex.peek(), value);
}

void tab::store(word windex, const tri & value)
{
    switch (tri::isWhat(windex)) {
    case tri::INT:{
	    int ix = tri::asInt(windex);
	    arr[ix] = value;
	    return;
	}
    case tri::SYM:{
	    int sz = dict.size();
	    for (int i = 0; i < sz; i += 2) {
		if (windex == dict[i].peek()) {
		    dict[i + 1] = value;
		    return;
		}
	    }
	    dict[sz] = tri().setSym(windex);
	    dict[sz + 1] = value;
	    return;
	}
    }
    CANT("tab::store() index nor INT nor SYM");
}


int atoi(char *s)
{
    int z = 0;
    while ('0' <= *s && *s <= '9') {
	z = z * 10 + *s - '0';
	++s;
    }
    return z;
}


tri& tri::setTCodeArrayFromString(const char* s, int* returnLastUsed, bool deeply )
{
if(verbose) DBG(stderr, "PARSING: %s\n", s );
    this->setNewTab();

    int n = 0;			// vec index

    int i = 0;
    int j = i;

    while (1) {

	while (0 < s[i] && s[i] < 33) {
	    ++i;
	}			// skip white
	j = i;

	if (!s[i])
	    break;		// stop on EOS

	j = i;
	while (s[j] < 0 || s[j] > 32) {
	    ++j;
	}			// find end of word

	tri token;
	token.setTCodeItemFromString(s, i, j);

	if (deeply) {
	    switch (token.peek()) {

	    case SYM1('{'): {
		int out;
		this->aPut(n, tri().setTCodeArrayFromString(s+j, &out)); // subarray
		j += out;
	    }
		break;

	    case SYM1('}'):
		goto finish;  // end subarray

	    default:
		this->aPut(n, token);
	    }
	} else {
		this->aPut(n, token);
	}

	++n;

	i = j;			// advance beyond that word
    }

finish:
    if (returnLastUsed) *returnLastUsed= j;
    return *this;
}

tri & tri::setTCodeItemFromString(const char *s, int i, int j)
{
    u1 *p = (u1 *) s;

    if ('0' <= s[i] && s[i] <= '9') {
	setInt(atoi((char *) s + i));
    } else if ( s[i]=='\"' ) {
	setNewTab();
	for (int k=1; i+k<j; k++) {
	    if ( s[i+k]=='\"' ) break;
	    aPut(k-1, (int)s[i+k]);
	}
    } else if (1 == j - i) {
	setSym(tri::SYM | p[i]);
    } else if (2 == j - i) {
	setSym(tri::SYM | (p[i] << 8) | p[i + 1]);
    } else if (3 == j - i) {
	setSym(tri::SYM | (p[i] << 16) | (p[i + 1] << 8) | p[i + 2]);
    } else {
	setSym(tri:: SYM | (p[i] << 24) | (p[i + 1] << 16) | (p[i + 2] << 8)
	       | p[i + 3]);
    }
    return *this;
}


TriFrame::TriFrame() : fflags(0), kp(), pc(0), bp(0), lobby() {

}


TriTerp::TriTerp() 
: home() 
, TCODE_MAGIC( tri().setSym( 't','c','o','d' ) )
{
    init();
}

void TriTerp::init() {
    // use this if zeros used instead of constructor
    home.setNewTab();
    //home.setTag( SYM4('h','o','m','e') );
    sp= & stack[0];
    f= & frames[0];
    f->lobby= this->home;   // defaults to the main lobby
    f->fflags = f->STOP;
    push( home );
    push( tri() );
}



void TriTerp::advanceFrame() 
{
    ++f;
    if ( f >= frames + MAX_FRAMES ) {
	Bomb("overflow in advanceFrame: too many frames");
    }
    f->lobby= (f-1)->lobby;   // defaults to the previous lobby
}

void TriTerp::retreatFrame() 
{
    --f;
    if ( f < frames ) {
	Bomb("underflow in retreatFrame: too many retreats");
    }
}


err TriTerp::testEval( tri tcodes, tri arg1, tri arg2 )
{
    // must be well-formed tcodes
    MUST( tcodes.isTab() );
    MUST( tcodes.asTab()->aSize() > 0 );
    MUST( tcodes.asTab()->aGet(0) == TCODE_MAGIC );

    sp= stack;
    push( tri('<','$','>') );  // begin-of-stack marker

    f->bp= sp;

    f->kp= tcodes;  // kode passed in as tcodes
    f->pc= 1;	    // skip magic slot

    push( arg1 );  // two hardwired args
    push( arg2 );

    return doLoop();
}

/////////////////////////////////////


#if DEBUG
void _tri(const tri* a) { say << *a; }
void _tab(const tab* a) { say << *a; }
void _vec(const vec* a) { say << *a; }
void _terp(const TriTerp* a) { say << *a; }


Sayer& Sayer::operator << (int a) {
	fprintf(stderr, "%d ", a);
	return *this;
}
Sayer& Sayer::operator << (const char* a) {
	fprintf(stderr, "%s ", a);
	return *this;
}
Sayer& Sayer::operator << (const char a) {
	fprintf(stderr, "%c ", a);
	return *this;
}
Sayer& Sayer::operator << (const tri& a) {
	tri t= NIL << a;
	int n= t.aSize();
	char* buf= new char[n+1];
	for (int i=0; i<n; i++) {
	    tri c= t.aGet(i);
	    if (c.isInt()) {
		int z= c.asInt();
		buf[i]= (32<=z && z<=126) ? z : '?';
	    } else {
		buf[i]= '#';
	    }
	}
	buf[n]= '\0';
	fprintf(stderr, "{%s} ", buf);
	delete [] buf;
	return *this;
}
Sayer& Sayer::operator << (const tab& x) {
    fprintf( stderr, "\n{ " );
    int n= x.aSize();
    for (int i=0; i<n; i++) {
	*this << x.aGet(i);
    }
    fprintf( stderr, " <[%d]> ", n );

    const vec& d= x.dict;
    n= d.size();
    for (int i=0; i<n; i++) {
	*this << d[i];
    }
    fprintf( stderr, " }\n", n );
    return *this;
}
Sayer& Sayer::operator << (const vec& a) {
    return *this;
}

Sayer& Sayer::operator << (const TriTerp& a) {
    a.dumpStack();
    return *this;
}

Sayer& Sayer::operator << (Expr* a) { a->print(); return *this; }
#endif




/////////////////////////////////////

err TriTerp::doTab( tri t ) { return NIL << "cannot execute TABLE as bytecode: " << t ; }
err TriTerp::doUnk( word w ) { return NIL << "cannot execute UNK as bytecode: " << w; }

inline void TriTerp::doInt( const tri& tcod ) { push(tcod); }

#define Kp	(f->kp)
#define Pc	(f->pc)
#define Bp	(f->bp)
#define Ksize() (ktab->aSize())
#define Kget(i) (ktab->aGet(i))
#define	Fflags	(f->fflags)
#define	Lobby	(f->lobby)


err TriTerp::doLoop()
{
    while (true) 
    {
	if ( Pc >= Kp.aSize() ) {
	    // falling off end of a vector.

	    if ( Fflags == f->LOOP ) {
		// go back and repeat loop
		Pc= 1;
		continue;

	    } else if ( Fflags == f->STOP ) {
		break;

	    } else if ( Fflags == f->FUNC ) {
		doReturn();
		continue;
	    } else {
		retreatFrame();
		continue;
	    }
	}

	const tri& tcod= Kp.aGet( Pc );
	++Pc;

#ifdef DEBUG
	if (verbose) 
	{
	    dumpStack();
	    char buf[99999]; 
	    tcod.toString(buf);
	    for ( int i=0; i<f-frames; i++) {
		DBG(stderr, "... " );
	    }
	    DBG(stderr, "%03d:%03d ... <%s>\n", f-frames, Pc-1, buf );
	}
#endif

	tri e;
	switch (tcod.isWhat()) {

	case tri::INT:	    doInt( tcod ); break;
	case tri::SYM:	    e= doSym( tcod ); break;
	case tri::TAB:	    e= doTab( tcod ); break;
	case tri::UNK:	    e= doUnk( tcod.peek() ); break;

	}

	if ( e.peek() != NIL_GUTS ) { return e << " pc=" << (Pc-1) << " in {" << Kp << "} "; }

    }

#ifdef DEBUG
    if (verbose) dumpStack();
#endif
    return tri(); // good result
}

inline tri TriTerp::pcImmediate() {
    const tri& z= Kp.aGet(Pc);
    ++Pc;
#ifdef DEBUG
    char buf[99999];
    if (verbose) {
	z.toString(buf);
	DBG( stderr, "        " );
	for ( int i=0; i<f-frames; i++) {
	    DBG(stderr, "... " );
	}
	DBG( stderr, "... %s\n", buf );
    }
#endif
    return z;
}

err TriTerp::doWhile() {
	tri clause= pcImmediate();

	if ( ! clause.isTab() ) return tri('w','h','i','!');  // must have an array!
	if ( tri('t','c','o','d') != clause.aGet(0) ) return tri('w','h','i','~');

	advanceFrame();
	f->fflags= f->LOOP;
	f->kp= clause;
	f->pc= 1;
	f->bp= f[-1].bp; // same bp

	return tri();
}

err TriTerp::doIf() {
	tri thenClause= pcImmediate();
	tri elseClause= pcImmediate();

	tri cond= pop();

	tri chosen= IsTrue(cond) ? thenClause: elseClause;

	if (chosen.isTab()) {

	    advanceFrame();
	    f->fflags= 0;
	    f->kp= chosen;
	    EQ( tri('t','c','o','d'), f->kp.aGet(0) );
	    f->pc= 1;
	    f->bp= f[-1].bp; // same bp
	} else if (chosen == NIL) {
	    push( NIL );
	} else {
	    return NIL << "strange chosen clause in if: " << chosen;
	}

	return tri();
}

err TriTerp::doLclInfo() {
	tri lcls= pcImmediate();
	int n= lcls.aSize();

#if 1
	if ( sp > f->bp + n ) {
		tri e= NIL 
		    << "too many (" << (sp - f->bp) 
		    << ") arguments passed to function needing " << n << " locals";
		if ( sp > f->bp ) e << " : " << f->bp[0];
		if ( sp > f->bp + 1 ) e << " : " << f->bp[1];
		return e;
	};
#else
	while ( sp > f->bp + n ) {
		IFD( DBG(stderr, "popping extra thing off of stack\n"); )
		pop();
	};
#endif

	while ( sp < f->bp + n ) {
		IFD( DBG(stderr, "pushing extra thing onto stack\n"); )
		push(NIL);
	};

	return NIL;
}

err TriTerp::doCall() {
	int nargs= pcImmediate().asInt();
	MUST( nargs >=0 );
	MUST( nargs <= (sp-Bp) );
	tri receiver= sp[-nargs];
	tri message= sp[-nargs+1];
	MUST( receiver.isTab() );
	advanceFrame();
	f->fflags= f->FUNC;
	f->kp= receiver.asTab()->recall(message);

	if ( ! f->kp.isTab() ) {
	    return NIL << "not-a-table" << f->kp << "receiver" << receiver << "message" << message;
	}

	if ( tri('t','c','o','d') != f->kp.aGet(0) ) 
	    return NIL << "not-a-tcod" << f->kp << "receiver" << receiver << "message" << message;

	f->pc= 1;
	f->bp= sp - nargs;
	return tri();
}

err TriTerp::doBreak()
{
    tri cond= pop();
    if ( ! cond.peek() ) return tri(); // dont break if cond is false

    while ( Fflags != f->STOP && Fflags != f->FUNC && Fflags != f->LOOP ) retreatFrame();
    if ( Fflags == f->STOP ) {
	return tri('b','r','k','$');   // should find LOOP before STOP
    }
    if ( Fflags == f->FUNC ) {
	return tri('b','r','k','@');   // should find LOOP before FUNC
    }
    
    retreatFrame();  // retreat from the LOOP

    return tri();
}

err TriTerp::doReturn()
{
    tri result= pop();

    while ( Fflags != f->STOP && Fflags != f->FUNC ) retreatFrame();
    if ( Fflags == f->STOP ) {
	return doStop();
    }
    
    retreatFrame();
    while ( sp > f[1].bp ) { *sp=tri(); --sp; }  // unwind stack

    push( result );

    return tri();
}

err TriTerp::doStop()
{
    Bomb("return in STOP frame");
    return tri('S','T','O','P');
}
err TriTerp::doAbort()
{
    say << "\n***** ABORTING(" << pop() << ")*****\n";
    dumpVec( home, "Home", 4 );
    dumpStack( );
    Bomb("ABORTING on `abor' tripe");
    abort();
    return tri('A','B','O','R');
}
err TriTerp::doPutc()
{
    tri x= pop();
    if ( x.isInt() ) {

	int prob= PutcStdout( (int) x.asInt() );

	return prob ?  NIL << "putc failed"  : NIL;

    } else {

	return NIL << "not-an-int in putc" << x;

    }
}
err TriTerp::doPuts()
{
    tri s= pop();
    tri e= PutsStdout( s );
    return e;
}


err PutsWithFunc( tri x, int(*func)(const char*,int) )
{
	if ( ! x.isTab() ) return NIL << "not-a-table in PutsStderr";
	int n= x.aSize();
	if (n==0) return NIL;

	char* buf= new char [ n ];
	for (int i=0; i<n; i++) {
		tri c= x.aGet( i );

		if ( ! c.isInt() ) {
			delete buf;
			return NIL << "not-an-int in table in PutsStderr";
		}
		buf[i]= (char) c.asInt();
	}
	int prob= func(buf,n);
	delete [] buf;
	return prob ? NIL << "error in PutsWithFunc" : NIL;
}
err PutsStdout( tri x ) { return PutsWithFunc( x, &PutsStdout ); }
err PutsStderr( tri x ) { return PutsWithFunc( x, &PutsStderr ); }


void dumpVec(tri x, const char* s, int depth)
{
    if ( depth==0 ) {
	PutsStderr( NIL << x << ' ' );
	return;
    }

    if ( ! x.isTab() ) {
	PutsStderr( NIL << x << ' ' );
	return;
    }
	
    PutsStderr(  s  ?  NIL<< "\n[[" << s << "]] "  :  NIL<< " { " ); 
    int n= x.aSize();
    if (IsPrintableAsciiObj(x))  {
	PutcStderr( '\"' );
	for (int i=0; i<n; i++) {
		PutcStderr( x.aGet(i).asInt() );
	}
	PutcStderr( '\"' );
    } else {
	for (int i=0; i<n; i++) {
		dumpVec( x.aGet(i), 0, depth-1 );
	}
    } 
    vec& d= x.asTab()->dict;
    int nd= d.size();
    if ( nd>0 ) PutsStderr( NIL<< " (*" << n << ":" << nd << "*) " );

    for (int i=0; i<nd; i++) {

	dumpVec( d[i], 0, depth-1 );
    }
    PutsStderr(  NIL << ( s ? "\n" : " } " ) );
}

void TriTerp::dumpStack() const
{
    for (int i=0; stack+i<sp; i++) {
	PutsStderr(NIL<< stack[i] << ' ');
    }
    PutsStderr( NIL<< "\n" );
}


err TriTerp::doSym( tri ww ) {
    tri e;
    word w= ww.peek();

    switch (w) {

    // binary ops
    case SYM1('+'):  { int y=pop().asSignedInt(); int x=pop().asSignedInt(); push( tri( x+y )); } break;
    case SYM1('-'):  { int y=pop().asSignedInt(); int x=pop().asSignedInt(); push( tri( x-y )); } break;
    case SYM1('*'):  { int y=pop().asSignedInt(); int x=pop().asSignedInt(); push( tri( x*y )); } break;
    case SYM1('/'):  { int y=pop().asSignedInt(); int x=pop().asSignedInt(); if (!y) goto div0; push( tri( x/y )); } break;
    case SYM1('%'):  { int y=pop().asSignedInt(); int x=pop().asSignedInt(); if (!y) goto div0; push( tri( x%y )); } break;
    case SYM1('&'):  { int y=pop().asInt(); int x=pop().asInt(); push( tri( x&y )); } break;
    case SYM1('|'):  { int y=pop().asInt(); int x=pop().asInt(); push( tri( x|y )); } break;
    case SYM1('^'):  { int y=pop().asInt(); int x=pop().asInt(); push( tri( x^y )); } break;

    case SYM2('=','='):  { word y=pop().asComparisonWord(); word x=pop().asComparisonWord(); push( tri( x==y )); } break;
    case SYM2('!','='):  { word y=pop().asComparisonWord(); word x=pop().asComparisonWord(); push( tri( x!=y )); } break;
    case SYM1('<'    ):  { word y=pop().asComparisonWord(); word x=pop().asComparisonWord(); push( tri( x< y )); } break;
    case SYM1('>'    ):  { word y=pop().asComparisonWord(); word x=pop().asComparisonWord(); push( tri( x> y )); } break;
    case SYM2('<','='):  { word y=pop().asComparisonWord(); word x=pop().asComparisonWord(); push( tri( x<=y )); } break;
    case SYM2('>','='):  { word y=pop().asComparisonWord(); word x=pop().asComparisonWord(); push( tri( x>=y )); } break;

    // array/table
    case SYM4('a','n','e','w'): { push( tri().setNewTab() ); } break;
    case SYM4('a','g','e','t'): { tri i=pop(); tab* tp=pop().asTab(); push( tp->recall(i) ); } break;
    case SYM4('a','p','u','t'): { tri x=pop(); tri i=pop(); tab* tp=pop().asTab(); tp->store(i,x); } break;
    case SYM4('a','l','e','n'): { tri x=pop(); if (x.isTab()) push(x.aSize()); else {e=x; goto alen;} } break;


    // special
    case SYM4('h','o','m','e'): { push( Lobby ); }  break;
    case SYM4('i','m','m','/'): { push( pcImmediate() ); } break;

    // base pointer relative
    case SYM4('s','t','o','/'): {       // ( x ) sto/ i ( )   Stores x to bp[i]
	int i= pcImmediate().asInt();
	tri x= pop();
	MUST( i>=0 );
	MUST( Bp+i < sp );
	f->bp[i]= x;
    }
    break;

    case SYM4('r','c','l','/'): {       //  ( ) rcl/ i ( x )    Recalls x from bp[i]
	int i= pcImmediate().asInt();
	MUST( i>=0 );
	MUST( Bp+i < sp );
	push( f->bp[i] );
    }
    break;

    case SYM3('n','i','l'):	{ push( NIL ); } break;

    case SYM3('p','o','p'):	{ pop(); } break;

    case SYM4('c','a','l','/'): { e= doCall(); } break;

    case SYM3('r','e','t'):     { e= doReturn(); } break;

    case SYM2('i','f'):		{ e= doIf(); } break;

    case SYM4('w','h','i','l'):	{ e= doWhile(); } break;

    case SYM4('l','o','o','p'):	{ e= doWhile(); } break;

    case SYM3('b','r','k'):	{ e= doBreak(); } break;

    case SYM4('p','u','t','c'):	{ e= doPutc(); } break;
    case SYM4('p','u','t','s'):	{ e= doPuts(); } break;

    case SYM4('l','c','l','/'):	{ e= doLclInfo(); } break;

    case SYM4('a','b','o','r'):	{ e= doAbort(); } break;

    default:
DBG(stderr, "unknown Tcod symbol 0x%08x\n", w );
	goto sunk;
    }

    return e;

sunk:
    Bad();
    return NIL << "unknown symbol :" << ww;
div0:
    return NIL << "division by zero";
alen:
    Bad();
    return NIL << "not an array: called alen on " << e;
}
